@@ -31,11 +31,9 @@
 #include <apt-pkg/algorithms.h>
 #include <apt-pkg/acquire-item.h>
 #include <apt-pkg/configuration.h>
-#include <apt-pkg/depcache.h>
 #include <apt-pkg/error.h>
 #include <apt-pkg/fileutl.h>
 #include <apt-pkg/init.h>
-#include <apt-pkg/pkgcachegen.h>
 #include <apt-pkg/pkgrecords.h>
 #include <apt-pkg/pkgsystem.h>
 #include <apt-pkg/policy.h>
@@ -46,12 +44,14 @@
 #include <sys/fcntl.h>
 
 // Qt includes
-#include <QtCore/QProcess>
+#include <QtCore/QFile>
 #include <QtDBus/QDBusConnection>
 #include <QtDBus/QDBusMessage>
 
 #define RAMFS_MAGIC     0x858458f6
 
+#include "../debfile.h"
+#include "debug.h"
 #include "qaptauthorization.h"
 #include "workeracquire.h"
 #include "workerinstallprogress.h"
@@ -61,36 +61,26 @@ QAptWorker::QAptWorker(int &argc, char **argv)
         , m_cache(0)
         , m_records(0)
         , m_questionBlock(0)
-        , m_questionResponse(QVariantMap())
         , m_systemLocked(false)
+        , m_initialized(false)
 {
     new QaptworkerAdaptor(this);
 
-    if (!QDBusConnection::systemBus().registerService("org.kubuntu.qaptworker")) {
+    if (!QDBusConnection::systemBus().registerService(QLatin1String("org.kubuntu.qaptworker"))) {
         // Another worker is already here, quit
         QTimer::singleShot(0, QCoreApplication::instance(), SLOT(quit()));
         return;
     }
 
-    if (!QDBusConnection::systemBus().registerObject("/", this)) {
+    if (!QDBusConnection::systemBus().registerObject(QLatin1String("/"), this)) {
         QTimer::singleShot(0, QCoreApplication::instance(), SLOT(quit()));
         return;
     }
 
-    QTimer::singleShot(10000, this, SLOT(quit()));
-    sleep(1);
-
-    m_acquireStatus = new WorkerAcquire(this);
-    connect(m_acquireStatus, SIGNAL(downloadProgress(int, int, int)),
-            this, SIGNAL(downloadProgress(int, int, int)));
-    connect(m_acquireStatus, SIGNAL(downloadMessage(int, const QString&)),
-            this, SIGNAL(downloadMessage(int, const QString&)));
-    connect(m_acquireStatus, SIGNAL(fetchError(int, const QVariantMap&)),
-            this, SIGNAL(errorOccurred(int, const QVariantMap&)));
-    connect(m_acquireStatus, SIGNAL(fetchWarning(int, const QVariantMap&)),
-            this, SIGNAL(warningOccurred(int, const QVariantMap&)));
-    connect(m_acquireStatus, SIGNAL(workerQuestion(int, const QVariantMap&)),
-            this, SIGNAL(questionOccurred(int, const QVariantMap&)));
+    m_timeout = new QTimer(this);
+    connect(m_timeout, SIGNAL(timeout()), this, SLOT(quit()));
+    m_timeout->setSingleShot(true);
+    m_timeout->start(30000);
 }
 
 QAptWorker::~QAptWorker()
@@ -104,16 +94,20 @@ void QAptWorker::setLocale(const QString &locale) const
     std::setlocale(LC_ALL, locale.toAscii());
 }
 
+void QAptWorker::setProxy(const QString &proxy) const
+{
+    setenv("http_proxy", proxy.toAscii(), 1);
+}
+
 bool QAptWorker::lockSystem()
 {
     if (m_systemLocked) {
         return true;
     }
 
-    bool systemLocked = _system->Lock();
-    m_systemLocked = systemLocked;
+    m_systemLocked = _system->Lock();
 
-    return systemLocked;
+    return m_systemLocked;
 }
 
 bool QAptWorker::unlockSystem()
@@ -130,42 +124,75 @@ bool QAptWorker::unlockSystem()
 
 bool QAptWorker::initializeApt()
 {
-    if (!pkgInitConfig(*_config)) {
-        return false;
-    }
-
-    if (!pkgInitSystem(*_config, _system)) {
-        return false;
-    }
-
-    delete m_cache;
-    m_cache = new QApt::Cache(this);
-    if (!m_cache->open()) {
-        QVariantMap details;
-        string message;
-        bool isError = _error->PopMessage(message);
-        if (isError) {
-            details["ErrorText"] = QString::fromStdString(message);
+    if (!m_initialized) {
+        if (!pkgInitConfig(*_config) || !pkgInitSystem(*_config, _system)) {
+            throwInitError();
+            return false;
         }
-        emit errorOccurred(QApt::InitError, details);
-        return false;
+
+        if (!m_cache) {
+            m_cache = new QApt::Cache(this);
+        }
+
+        m_initialized = true;
     }
 
-    pkgDepCache *depCache = m_cache->depCache();
-
-    delete m_records;
-    m_records = new pkgRecords(*depCache);
+    reloadCache();
 
     return true;
 }
 
+void QAptWorker::reloadCache()
+{
+    if (!m_cache->open()) {
+        aptDebug() << "Cache didn't open";
+        throwInitError();
+        return;
+    }
+
+    delete m_records;
+    m_records = new pkgRecords(*m_cache->depCache());
+}
+
+void QAptWorker::throwInitError()
+{
+    QVariantMap details;
+    string message;
+    bool isError = _error->PopMessage(message);
+    if (isError) {
+        details[QLatin1String("FromWorker")] = true;
+        details[QLatin1String("ErrorText")] = QString::fromStdString(message);
+    }
+
+    errorOccurred(QApt::InitError, details);
+}
+
+void QAptWorker::initializeStatusWatcher()
+{
+    m_acquireStatus = new WorkerAcquire(this);
+    connect(m_acquireStatus, SIGNAL(downloadProgress(int, int, int)),
+            this, SIGNAL(downloadProgress(int, int, int)));
+    connect(m_acquireStatus, SIGNAL(packageDownloadProgress(const QString&, int, const QString&, double, int)),
+            this, SIGNAL(packageDownloadProgress(const QString&, int, const QString&, double, int)));
+    connect(m_acquireStatus, SIGNAL(downloadMessage(int, const QString&)),
+            this, SIGNAL(downloadMessage(int, const QString&)));
+    connect(m_acquireStatus, SIGNAL(fetchError(int, const QVariantMap&)),
+            this, SIGNAL(errorOccurred(int, const QVariantMap&)));
+    connect(m_acquireStatus, SIGNAL(fetchWarning(int, const QVariantMap&)),
+            this, SIGNAL(warningOccurred(int, const QVariantMap&)));
+    connect(m_acquireStatus, SIGNAL(workerQuestion(int, const QVariantMap&)),
+            this, SIGNAL(questionOccurred(int, const QVariantMap&)));
+}
+
 void QAptWorker::updateCache()
 {
-    if (!QApt::Auth::authorize("org.kubuntu.qaptworker.updateCache", message().service())) {
+    if (!QApt::Auth::authorize(QLatin1String("org.kubuntu.qaptworker.updateCache"), message().service())) {
         emit errorOccurred(QApt::AuthError, QVariantMap());
         return;
     }
 
+    m_timeout->stop();
+
     emit workerStarted();
     emit workerEvent(QApt::CacheUpdateStarted);
 
@@ -184,16 +211,18 @@ void QAptWorker::updateCache()
         }
     }
 
-    pkgAcquire fetcher(m_acquireStatus);
+    initializeStatusWatcher();
+    pkgAcquire fetcher;
+    fetcher.Setup(m_acquireStatus);
     // Populate it with the source selection and get all Indexes
     // (GetAll=true)
-    if (m_cache->list()->GetIndexes(&fetcher,true) == false) {
+    if (!m_cache->list()->GetIndexes(&fetcher,true)) {
         emit workerFinished(false);
         return;
     }
 
     // do the work
-    if (_config->FindB("APT::Get::Download",true) == true) {
+    if (_config->FindB("APT::Get::Download",true)) {
         bool result = ListUpdate(*m_acquireStatus, *m_cache->list());
         emit workerFinished(result);
     } else {
@@ -202,25 +231,25 @@ void QAptWorker::updateCache()
     }
 
     emit workerEvent(QApt::CacheUpdateFinished);
+    m_timeout->start();
 }
 
 void QAptWorker::cancelDownload()
 {
-    m_acquireStatus->requestCancel();
+    if (m_acquireStatus) {
+        m_acquireStatus->requestCancel();
+    }
 }
 
 void QAptWorker::commitChanges(QMap<QString, QVariant> instructionsList)
 {
-    if (!QApt::Auth::authorize("org.kubuntu.qaptworker.commitChanges", message().service())) {
-        emit errorOccurred(QApt::AuthError, QVariantMap());
-        return;
-    }
-
     if (!initializeApt()) {
         emit workerFinished(false);
         return;
     }
 
+    m_timeout->stop();
+
     QVariantMap versionList;
 
     // Parse out the argument list and mark packages for operations
@@ -234,16 +263,19 @@ void QAptWorker::commitChanges(QMap<QString, QVariant> instructionsList)
         QString packageString = mapIter.key();
         QString version;
 
-        if (packageString.contains(',')) {
-            QStringList split = packageString.split(',');
+        // Check if a version is specified
+        if (packageString.contains(QLatin1Char(','))) {
+            QStringList split = packageString.split(QLatin1Char(','));
             iter = m_cache->depCache()->FindPkg(split.at(0).toStdString());
             version = split.at(1);
         } else {
             iter = m_cache->depCache()->FindPkg(packageString.toStdString());
         }
+
+        // Somehow the package searched for doesn't exist.
         if (iter == 0) {
             QVariantMap args;
-            args["NotFoundString"] = packageString;
+            args[QLatin1String("NotFoundString")] = packageString;
             emit errorOccurred(QApt::NotFoundError, args);
             emit workerFinished(false);
             return;
@@ -258,7 +290,17 @@ void QAptWorker::commitChanges(QMap<QString, QVariant> instructionsList)
                 m_cache->depCache()->MarkKeep(iter, false);
                 m_cache->depCache()->SetReInstall(iter, false);
                 break;
-            case QApt::Package::ToUpgrade:
+            case QApt::Package::ToUpgrade: {
+                bool fromUser = !(State.Flags & pkgCache::Flag::Auto);
+                m_cache->depCache()->MarkInstall(iter, true, 0, fromUser);
+                if (!State.Install() || m_cache->depCache()->BrokenCount() > 0) {
+                    pkgProblemResolver Fix(m_cache->depCache());
+                    Fix.Clear(iter);
+                    Fix.Protect(iter);
+                    Fix.Resolve(true);
+                }
+                break;
+            }
             case QApt::Package::ToInstall:
                 m_cache->depCache()->MarkInstall(iter, true);
                 if (!State.Install() || m_cache->depCache()->BrokenCount() > 0) {
@@ -315,7 +357,9 @@ void QAptWorker::commitChanges(QMap<QString, QVariant> instructionsList)
         }
     }
 
-    pkgAcquire fetcher(m_acquireStatus);
+    initializeStatusWatcher();
+    pkgAcquire fetcher;
+    fetcher.Setup(m_acquireStatus);
 
     pkgPackageManager *packageManager;
     packageManager = _system->CreatePM(m_cache->depCache());
@@ -336,12 +380,12 @@ void QAptWorker::commitChanges(QMap<QString, QVariant> instructionsList)
         emit warningOccurred(QApt::SizeMismatchWarning, QVariantMap());
     }
 
-    /* Check for enough free space */
+    // Check for enough free space
     struct statvfs Buf;
     string OutputDir = _config->FindDir("Dir::Cache::Archives");
     if (statvfs(OutputDir.c_str(),&Buf) != 0) {
         QVariantMap args;
-        args["DirectoryString"] = QString::fromStdString(OutputDir.c_str());
+        args[QLatin1String("DirectoryString")] = QLatin1String(OutputDir.c_str());
         emit errorOccurred(QApt::DiskSpaceError, args);
         emit workerFinished(false);
         return;
@@ -353,7 +397,7 @@ void QAptWorker::commitChanges(QMap<QString, QVariant> instructionsList)
             unsigned(Stat.f_type)            != RAMFS_MAGIC)
         {
             QVariantMap args;
-            args["DirectoryString"] = QString::fromStdString(OutputDir.c_str());
+            args[QLatin1String("DirectoryString")] = QLatin1String(OutputDir.c_str());
             emit errorOccurred(QApt::DiskSpaceError, args);
             emit workerFinished(false);
             return;
@@ -371,7 +415,7 @@ void QAptWorker::commitChanges(QMap<QString, QVariant> instructionsList)
 
     if (!untrustedPackages.isEmpty()) {
         QVariantMap args;
-        args["UntrustedItems"] = untrustedPackages;
+        args[QLatin1String("UntrustedItems")] = untrustedPackages;
 
         if (_config->FindB("APT::Get::AllowUnauthenticated", true) == true) {
             // Ask if the user really wants to install untrusted packages, if
@@ -383,7 +427,8 @@ void QAptWorker::commitChanges(QMap<QString, QVariant> instructionsList)
             emit questionOccurred(QApt::InstallUntrusted, args);
             m_questionBlock->exec();
 
-            bool m_installUntrusted = m_questionResponse["InstallUntrusted"].toBool();
+            // m_questionBlock will block function until we get a response in m_installUntrusted
+            bool m_installUntrusted = m_questionResponse[QLatin1String("InstallUntrusted")].toBool();
             if(!m_installUntrusted) {
                 m_questionResponse = QVariantMap(); //Reset for next question
                 emit workerFinished(false);
@@ -399,6 +444,12 @@ void QAptWorker::commitChanges(QMap<QString, QVariant> instructionsList)
         }
     }
 
+    if (!QApt::Auth::authorize(QLatin1String("org.kubuntu.qaptworker.commitChanges"), message().service())) {
+        emit errorOccurred(QApt::AuthError, QVariantMap());
+        emit workerFinished(false);
+        return;
+    }
+
     emit workerEvent(QApt::PackageDownloadStarted);
 
     if (fetcher.Run() != pkgAcquire::Continue) {
@@ -433,6 +484,151 @@ void QAptWorker::commitChanges(QMap<QString, QVariant> instructionsList)
 
     delete installProgress;
     installProgress = 0;
+
+    m_timeout->start();
+}
+
+void QAptWorker::downloadArchives(const QStringList &packageStrings, const QString &dest)
+{
+    if (!initializeApt()) {
+        emit workerFinished(false);
+        return;
+    }
+
+    m_timeout->stop();
+
+    emit workerStarted();
+
+    initializeStatusWatcher();
+    pkgAcquire fetcher;
+    fetcher.Setup(m_acquireStatus);
+
+    pkgIndexFile *index;
+
+    foreach (const QString &packageString, packageStrings) {
+        pkgCache::PkgIterator iter = m_cache->depCache()->FindPkg(packageString.toStdString());
+
+        if (!iter) {
+            // Package not found
+            continue;
+        }
+
+        pkgCache::VerIterator ver = (*m_cache->depCache()).GetCandidateVer(iter);
+
+        if (!ver || !ver.Downloadable() || !ver.Arch()) {
+            // Virtual package or not downloadable or broken
+            continue;
+        }
+
+        // Obtain package info
+        pkgCache::VerFileIterator vf = ver.FileList();
+        pkgRecords::Parser &rec = m_records->Lookup(ver.FileList());
+
+        // Try to cross match against the source list
+        if (!m_cache->list()->FindIndex(vf.File(), index)) {
+            continue;
+        }
+
+        string fileName = rec.FileName();
+        string MD5sum = rec.MD5Hash();
+
+        if (fileName.empty()) {
+            //TODO: Error
+            return;
+        }
+
+        new pkgAcqFile(&fetcher,
+                       index->ArchiveURI(fileName),
+                       MD5sum,
+                       ver->Size,
+                       index->ArchiveInfo(ver),
+                       ver.ParentPkg().Name(),
+                       dest.toStdString(), "");
+    }
+
+    emit workerEvent(QApt::PackageDownloadStarted);
+
+    if (fetcher.Run() != pkgAcquire::Continue) {
+        // Our fetcher will report errors for itself, but we have to send the
+        // finished signal
+        emit workerFinished(false);
+        return;
+    }
+
+    m_timeout->start();
+
+    emit workerEvent(QApt::PackageDownloadFinished);
+    emit workerFinished(true);
+}
+
+void QAptWorker::installDebFile(const QString &fileName)
+{
+    if (!initializeApt()) {
+        emit workerFinished(false);
+        return;
+    }
+    m_timeout->stop();
+
+    emit workerStarted();
+
+    QApt::DebFile deb(fileName);
+
+    QString arch = deb.architecture();
+
+    if (arch != QLatin1String("all") &&
+        arch != QString::fromStdString(_config->Find("APT::Architecture", ""))) {
+        // TODO report what arch was provided vs needed
+        aptDebug() << arch << QString::fromStdString(_config->Find("APT::Architecture", ""));
+        emit errorOccurred(QApt::WrongArchError, QVariantMap());
+        emit workerFinished(false);
+        return;
+    }
+
+    if (!QApt::Auth::authorize(QLatin1String("org.kubuntu.qaptworker.commitChanges"), message().service())) {
+        emit errorOccurred(QApt::AuthError, QVariantMap());
+        emit workerFinished(false);
+        return;
+    }
+
+    m_dpkgProcess = new QProcess(this);
+    QString program = QLatin1Literal("dpkg") %
+                      QLatin1Literal(" -i ") % fileName;
+    setenv("PATH", "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin", 1);
+    setenv("DEBIAN_FRONTEND", "passthrough", 1);
+    setenv("DEBCONF_PIPE", "/tmp/qapt-sock", 1);
+    m_dpkgProcess->start(program);
+    connect(m_dpkgProcess, SIGNAL(started()), this, SLOT(dpkgStarted()));
+    connect(m_dpkgProcess, SIGNAL(readyRead()), this, SLOT(updateDpkgProgress()));
+    connect(m_dpkgProcess, SIGNAL(finished(int, QProcess::ExitStatus)),
+            this, SLOT(dpkgFinished(int, QProcess::ExitStatus)));
+
+}
+
+void QAptWorker::dpkgStarted()
+{
+    emit workerEvent(QApt::DebInstallStarted);
+}
+
+void QAptWorker::updateDpkgProgress()
+{
+    QString str = m_dpkgProcess->readLine();
+
+    emit debInstallMessage(str);
+}
+
+void QAptWorker::dpkgFinished(int exitCode, QProcess::ExitStatus exitStatus)
+{
+    Q_UNUSED(exitCode);
+
+    if (!exitStatus) {
+        emit workerEvent(QApt::DebInstallFinished);
+    }
+
+    m_timeout->start();
+
+    emit workerFinished(!exitStatus);
+    delete m_dpkgProcess;
+    m_dpkgProcess = 0;
 }
 
 void QAptWorker::answerWorkerQuestion(const QVariantMap &response)
@@ -450,13 +646,72 @@ void QAptWorker::setAnswer(const QVariantMap &answer)
 
 void QAptWorker::updateXapianIndex()
 {
+    m_timeout->stop();
     emit workerEvent(QApt::XapianUpdateStarted);
-    m_xapianProc = new QProcess(this);
-    QString cmd = "/usr/bin/nice /usr/bin/ionice -c3 "
-                  "/usr/sbin/update-apt-xapian-index "
-                  "--update";
-    m_xapianProc->start(cmd);
 
-    m_xapianProc->waitForFinished();
+
+    QDBusMessage m = QDBusMessage::createMethodCall(QLatin1String("org.debian.AptXapianIndex"),
+                                                    QLatin1String("/"),
+                                                    QLatin1String("org.debian.AptXapianIndex"),
+                                                    QLatin1String("update_async"));
+    QVariantList dbusArgs;
+
+    dbusArgs << /*force*/ true << /*update_only*/ true;
+    m.setArguments(dbusArgs);
+    QDBusConnection::systemBus().send(m);
+
+    QDBusConnection::systemBus().connect(QLatin1String("org.debian.AptXapianIndex"), QLatin1String("/"), QLatin1String("org.debian.AptXapianIndex"),
+                                         QLatin1String("UpdateProgress"), this, SIGNAL(xapianUpdateProgress(int)));
+    QDBusConnection::systemBus().connect(QLatin1String("org.debian.AptXapianIndex"), QLatin1String("/"), QLatin1String("org.debian.AptXapianIndex"),
+                                         QLatin1String("UpdateFinished"), this, SLOT(xapianUpdateFinished(bool)));
+}
+
+void QAptWorker::xapianUpdateFinished(bool result)
+{
+    Q_UNUSED(result);
+    QDBusConnection::systemBus().disconnect(QLatin1String("org.debian.AptXapianIndex"), QLatin1String("/"), QLatin1String("org.debian.AptXapianIndex"),
+                                            QLatin1String("UpdateProgress"), this, SIGNAL(xapianUpdateProgress(int)));
+    QDBusConnection::systemBus().disconnect(QLatin1String("org.debian.AptXapianIndex"), QLatin1String("/"), QLatin1String("org.debian.AptXapianIndex"),
+                                            QLatin1String("UpdateFinished"), this, SLOT(xapianUpdateFinished(bool)));
+
     emit workerEvent(QApt::XapianUpdateFinished);
+    m_timeout->start();
+}
+
+bool QAptWorker::writeFileToDisk(const QString &contents, const QString &path)
+{
+    if (!QApt::Auth::authorize(QLatin1String("org.kubuntu.qaptworker.writeFileToDisk"), message().service())) {
+        emit errorOccurred(QApt::AuthError, QVariantMap());
+        return false;
+    }
+
+    QFile file(path);
+
+    if (file.open(QIODevice::WriteOnly | QIODevice::Text)) {
+        file.write(contents.toStdString().c_str());
+
+        return true;
+    }
+
+    return false;
+}
+
+bool QAptWorker::copyArchiveToCache(const QString &archivePath)
+{
+    if (!QApt::Auth::authorize(QLatin1String("org.kubuntu.qaptworker.writeFileToDisk"), message().service())) {
+        emit errorOccurred(QApt::AuthError, QVariantMap());
+        return false;
+    }
+
+    initializeApt();
+    QString cachePath = QString::fromStdString(_config->FindDir("Dir::Cache::Archives"));
+    // Filename
+    cachePath += archivePath.right(archivePath.size() - archivePath.lastIndexOf('/'));
+
+    if (QFile::exists(cachePath)) {
+        // Already copied
+        return true;
+    }
+
+    return QFile::copy(archivePath, cachePath);
 }
